﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.environment;

// スレッド別環境変数クラス
private import std.path;

private import pes.system.singleton;
private import pes.system.semaphore;
private import common.task;

import common.configuration;
import common.keytable;
import common.nodetable;
import common.connectiontable;
import common.cacheheadertable;
import common.workertable;
import common.downloadlist;

import worker.downloader;
import worker.uploader;
import worker.cacheencoder;
import worker.cachedecoder;

class Environment
{
protected:
	bool m_initialize;
	// 設定類
	Configuration		m_configuration;
	
	// テーブル類
	KeyTable 					m_key_table;			// キーテーブル
	NodeTable					m_node_table;			// ノードテーブル
	ConnectionTable				m_upstream_connection;	// 上流検索用コネクション
	ConnectionTable				m_downstream_connection;// 下流検索用コネクション
	ConnectionTable				m_ul_data_connection;	// ULデータ転送用コネクション
	ConnectionTable				m_dl_data_connection;	// DLデータ転送用コネクション
	CacheHeaderTable			m_cache_header;			// キャッシュヘッダ
	WorkerTable!(Uploader)		m_uploader;				// アップローダー
	WorkerTable!(Downloader)	m_downloader;			// ダウンローダー
	WorkerTable!(CacheEncoder)	m_encoder;				// エンコーダー
	WorkerTable!(CacheDecoder)	m_decoder;				// デコーダー
	DownloadList				m_download_list;		// ダウンロードリスト
	
public:
	void initialize(SemaphoreResource resource, Configuration configration)
	{
		m_initialize 			= true;
		m_key_table				= new KeyTable(resource, SemaphoreId.KEY);
		m_configuration 		= configration;
		m_node_table			= new NodeTable(resource, SemaphoreId.NODE);
		m_upstream_connection	= new ConnectionTable(resource, SemaphoreId.UPSTREAM);
		m_downstream_connection	= new ConnectionTable(resource, SemaphoreId.DOWNSTREAM);
		m_ul_data_connection	= new ConnectionTable(resource,  SemaphoreId.UL_DATA);
		m_dl_data_connection	= new ConnectionTable(resource, SemaphoreId.DL_DATA);
		m_uploader				= new WorkerTable!(Uploader)(resource, SemaphoreId.UPLOADER);
		m_downloader			= new WorkerTable!(Downloader)(resource, SemaphoreId.DOWNLOADER);
		m_encoder				= new WorkerTable!(CacheEncoder)(resource, SemaphoreId.ENCODER);
		m_decoder				= new WorkerTable!(CacheDecoder)(resource, SemaphoreId.DECODER);
		m_download_list			= new DownloadList(resource, SemaphoreId.DOWNLOAD_ITEM);
		
		m_cache_header			= new CacheHeaderTable(m_configuration.cacheIndex, resource, SemaphoreId.CACHE_INDEX);
	}
	
	void terminate()
	{
		if (m_initialize == false) {
			return;
		}
		m_initialize = false;
		
		m_uploader.clear();
		m_downloader.clear();
		m_encoder.clear();
		m_decoder.clear();
		m_key_table.close();
		m_node_table.clear();
		m_upstream_connection.clear();
		m_downstream_connection.clear();
		m_ul_data_connection.clear();
		m_dl_data_connection.clear();
		m_cache_header.close();
		m_download_list.clear();
		
		m_configuration 		= null;
		m_key_table				= null;
		m_node_table			= null;
		m_upstream_connection	= null;
		m_downstream_connection	= null;
		m_ul_data_connection	= null;
		m_dl_data_connection	= null;
		m_cache_header			= null;
		m_uploader				= null;
		m_downloader			= null;
		m_encoder				= null;
		m_decoder				= null;
		m_download_list			= null;
	}
	
	bool isVaild()
	{
		return m_initialize;
	}
	
	// getter
	
	Configuration configuration()
	{
		return m_configuration;
	}
	
	KeyTable keyTable()
	{
		return m_key_table;
	}
	
	NodeTable nodeTable()
	{
		return m_node_table;
	}
	
	ConnectionTable upstreamConnection()
	{
		return m_upstream_connection;
	}
	
	ConnectionTable downstreamConnection()
	{
		return m_downstream_connection;
	}
	
	ConnectionTable dlDataConnection()
	{
		return m_dl_data_connection;
	}
	
	ConnectionTable ulDataConnection()
	{
		return m_ul_data_connection;
	}
	
	CacheHeaderTable cacheHeader()
	{
		return m_cache_header;
	}
	
	WorkerTable!(CacheDecoder) decoder()
	{
		return m_decoder;
	}
	
	WorkerTable!(CacheEncoder) encoder()
	{
		return m_encoder;
	}
	
	WorkerTable!(Uploader) uploader()
	{
		return m_uploader;
	}
	
	WorkerTable!(Downloader) downloader()
	{
		return m_downloader;
	}
	
	DownloadList downloadList()
	{
		return m_download_list;
	}
}

ThreadSingleton!(Environment) g_environ;

static this()
{
	g_environ = new ThreadSingleton!(Environment);
}

public Environment getEnvironment()
{
	return g_environ.get();
}


