﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.net.winny.connection;

// Winny コネクションクラス



// remark; receiveする前にpollするのはそのうちやめれるようにする
// remark; 通信バッファにmalloc/freeをおこなう
// remark; 1度コネクトしてクローズするともう使えない

private import std.random;
private import std.socket;
private import std.string;
private import std.gc;
private import std.c.stdlib;
private import std.c.time;
private import std.outofmemory;

private import pes.rc4;
private import pes.random;
private import pes.queue;
private import pes.system.delay;
private import pes.net.util;
private import pes.net.winny.command;

private import std.stdio;

class ConnectionTimeOut: SocketException
{
	this()
	{
		super("コネクションタイムアウト");
	}
}

class ReceiveError: SocketException
{
	this(int errcd)
	{
		super("受信エラー", errcd);
	}
	this()
	{
		super("受信エラー", 0);
	}
}

class SendError: SocketException
{
	this(int errcd)
	{
		super("送信エラー", errcd);
	}
}

class ConnectionClosed: SocketException
{
	this()
	{
		super("切断エラー");
	}
}

class DecryptError: SocketException
{
	this(uint len)
	{
		super(format("受信異常 %sbytes", len));
	}
}

version(linux)
{
	private import std.c.linux.linux;
}

class NyConnection
{
	static uint s_malloc_queue;
protected:
	const uint				BUFFER_SIZE		= 0x100000;
	const uint				BLOCK_MAX		= 0x90000;
	const uint				SPEED_SEC		= 30;
	const uint				RETRY_MAX		= 3;
	
	struct Register
	{
		enum ReceivePhase {
			ZERO	= 0,
			FIRST 	= 1,
			SECOND	= 2,
		}
		uint tail, command_offset, packlen;
		int recvlen;
		ReceivePhase phase;
		
		void save(uint tail, uint command_offset, uint packlen, int recvlen, ReceivePhase phase)
		{
			this.tail = tail;
			this.command_offset = command_offset;
			this.packlen		= packlen;
			this.recvlen		= recvlen;
			this.phase			= phase;
		}
		
		void load(out uint tail, out uint command_offset,
					out uint packlen, out int recvlen, out ReceivePhase phase)
		{
			tail 	= this.tail;
			command_offset 	= this.command_offset;
			packlen			= this.packlen;
			recvlen			= this.recvlen;
			phase			= this.phase;
		}
		
		void clear()
		{
			tail 	= 0;
			command_offset 	= 0;
			packlen			= 0;
			recvlen			= 0;
			phase			= ReceivePhase.ZERO;
		}
	}
	Register				m_register;
	Exception				m_exception;
	InternetAddress			m_inetaddr;
	Socket					m_socket;
	SocketSet				m_fdset;
	RC4						m_recv_rc4key;
	RC4						m_send_rc4key;
	timeval					m_timeval;
	timeval					m_lasttimeval;
	bool					m_is_listener;
	Queue!(NyRawCommand)	m_command_queue;
	bool					m_is_active;
	ubyte[]					m_buffer;
	void 					*m_buffer_ptr;
	uint					m_send_rate_max		= uint.max;
	uint					m_send_rate_delay;
	uint					m_recv_rate_max		= uint.max;
	uint					m_recv_rate_delay;
	
	/* 通信管理関連	*/
	uint					m_recv_size_sec;	// 受信速度(B/S)
	time_t					m_last_recv_time;	// 最終受信時間
	uint					m_send_size_sec;	// 送信速度(B/S)
	time_t					m_last_send_time;	// 最終送信時間
	time_t					m_start_time;
	
	void connectionInitialize()
	{
//		//delete m_command_queue;
//		m_command_queue = null;
//		m_command_queue = new Queue!(NyRawCommand);
//		
//		//delete m_socket;
		if (m_socket !is null) {
			m_socket.close();
		}
//		m_socket = null;
//		
//		//delete m_inetaddr;
//		m_inetaddr = null;
//		
//		m_is_active = false;
	}
	
	// 通信レート制御
	void rateControl(inout uint rate_max, inout uint delay)
	{
		if (rate() > rate_max) {
			if (delay != 0) {
				--delay;
			}
		} else {
			++delay;
		}
		if (delay) {
			delayTask(delay * 10); // 10ms単位
		}
	}
	
	void upstreamRateControl()
	{
		rateControl(m_send_rate_max, m_send_rate_delay);
	}
	
	void downstreamRateControl()
	{
		rateControl(m_recv_rate_max, m_recv_rate_delay);
	}
	
	void send(ubyte[] packet)
	{
		time_t 	cur;
		int  	sendlen;
		uint	packet_size = packet.length;
		uint	pos;
		uint 	block_count		= packet_size / 0x2000;
		uint 	remainder		= packet_size % 0x2000;
		
		packet = m_send_rc4key.crypt(packet);
		
		for (pos = 0; pos < block_count; ++pos) {
//			upstreamRateControl();	// 上がり制御
			
			if (! m_socket.blocking) {
				poll(Poll.WRITEABLE);
			}
			sendlen = m_socket.send(packet[pos * 0x2000 .. ((pos + 1) * 0x2000)]);
			if (sendlen != 0x2000) {
				throw new SendError(sendlen);
			}
			
			// 通信速度計算
			cur = time(null);
			if (m_last_send_time + SPEED_SEC > cur) {
				m_send_size_sec += 0x2000;
			} else {
				m_send_size_sec = 0x2000;
				m_last_send_time = cur;
			}
		}
		if (remainder != 0) {
			if (! m_socket.blocking) {
				poll(Poll.WRITEABLE);
			}
			sendlen = m_socket.send(packet[pos * 0x2000 .. (pos * 0x2000) + remainder]);
			if (sendlen < 0) {
				throw new SendError(sendlen);
			}
			
			// 通信速度計算
			cur = time(null);
			if (m_last_send_time + SPEED_SEC > cur) {
				m_send_size_sec += remainder;
			} else {
				m_send_size_sec = remainder;
				m_last_send_time = cur;
			}
		}
	}
	
	int receive(ubyte[] buff, bool decrypt = true)
	{
		int recvlen;
		time_t cur;
		
		if (! isActive) {
			throw new ConnectionClosed();
		}
		if (! m_socket.blocking) {
			poll(Poll.__READABLE);
		}
		
		
//		downstreamRateControl();	// 下り制御
		
		recvlen = m_socket.receive(buff);
		if (recvlen < 0) {
			throw new ReceiveError(recvlen);
		} else if (recvlen == 0) {
			throw new ConnectionClosed();
		}
		
		// 通信速度計算
		cur = time(null);
		if (m_last_recv_time + SPEED_SEC > cur) {
			m_recv_size_sec += recvlen;
		} else {
			m_recv_size_sec = recvlen;
			m_last_recv_time = cur;
		}
		
		if (decrypt) {
			m_recv_rc4key.crypt(buff[0 .. recvlen]);
		}
		
		return recvlen;
	}
	
	NyProtocolHeader authorize()
	{
		NyProtocolHeader	header			= new NyProtocolHeader;
		NyLowVersion		close_low		= new NyLowVersion;
		NyRawCommand		command			= new NyRawCommand;
		ubyte[]				init_block		= generateRandomBytes(6);
		int					tryover, recvlen;
		
		m_send_rc4key.key(init_block[2 .. 6]);
		
		header.major(NyProtocolHeader.WINNY_MAJOR);
		header.minor(NyProtocolHeader.WINNY_MINOR);
		
		// 通信開始ブロック[ダミー:通信鍵] + 旧バージョン切断要求 + Winnyプロトコルヘッダ
		m_socket.send(init_block ~ (m_send_rc4key.crypt(close_low.pack ~ header.pack)));
		
		// 通信鍵のスクランブル
		m_send_rc4key.key(toStep2Key(init_block[2 .. 6]));
		
		// 接続先通信鍵の取得
		recvlen = receive(init_block, false);
		if (recvlen != init_block.length) {
			throw new NyBadCommand("通信暗号鍵の受信に失敗");
		}
		m_recv_rc4key.key(init_block[2 .. 6]);
		
//		writef("--- 1\n");
		// 接続先のWinnyプロトコルヘッダを受信するまでループ // 今のところ最大２回チャンレンジ
		tryover = 0;
		do {
//			writef("--- 2-%s\n", tryover);
			if (tryover > 2) {
				throw new NyBadCommand("Winnyプロトコルヘッダ受信に失敗(トライオーバー)");
			} ++tryover;
			command = receiveOneCommand();
		} while (command.commandCode != NyCommandCode.PROTOCOL_HEADER);
//		writef("--- 3-%s\n", command.toString);
		
		m_recv_rc4key.key(toStep2Key(init_block[2 .. 6]));
		
		return new NyProtocolHeader(command);
	}
	
	
	// コマンド補給
	void replenish()
	{
		NyRawCommand command;
		
		int recvlen = 0;
		uint packlen;
		uint tail = 0, command_offset = 0;
		Register.ReceivePhase phase;
		
		m_register.load(tail, command_offset, packlen, recvlen, phase);
		
		switch (phase) {
		case Register.ReceivePhase.FIRST:
			goto RECEIVE_PHASE_FIRST;
		case Register.ReceivePhase.SECOND:
			goto RECEIVE_PHASE_SECOND;
		default:
			break;
		}
		
		while (true) {
			// 最低length分受信する
			if (recvlen < CommandLength.sizeof) {
				while (recvlen < CommandLength.sizeof) {
			RECEIVE_PHASE_FIRST:
					try {
						recvlen += receive(m_buffer[command_offset + recvlen .. length]);
					} catch (ConnectionTimeOut e) {
						// 保存
						m_register.save(tail, command_offset, packlen, recvlen,
																Register.ReceivePhase.FIRST);
						throw e;
					}
				}
				tail += recvlen;
			}
			
			// lengthを取得
			packlen = *cast(uint *)(m_buffer.ptr + command_offset);
			if (packlen > BLOCK_MAX) {
				throw new DecryptError(packlen);
			}
			
			if (packlen == recvlen - CommandLength.sizeof) {
				/* 受信長と一致する場合	*/
				
				// キューに貯める
				command = new NyRawCommand(m_buffer[command_offset .. tail].dup);
				m_command_queue.push(command);
				
				m_register.clear();
				return;
			}
			
			if (packlen > recvlen - CommandLength.sizeof) {
				/* 受信長のほうが小さい場合	*/
				
				if (m_buffer.length - command_offset < packlen + CommandLength.sizeof) {
					// reallocate
					uint sz 	= command_offset + packlen + CommandLength.sizeof;
					ubyte *p 	= cast(ubyte *)realloc(m_buffer_ptr, sz);
					
					if (!p) {
						throw new OutOfMemoryException;
					}
					m_buffer	= p[0 .. sz];
					m_buffer_ptr = p;
				}
				
				// packlen分受信する
				do {
			RECEIVE_PHASE_SECOND:
					try {
						recvlen += receive(m_buffer[command_offset + recvlen 
													.. command_offset + packlen + CommandLength.sizeof]);
					} catch (ConnectionTimeOut e) {
						// 保存
						m_register.save(tail, command_offset, packlen, recvlen,
																Register.ReceivePhase.SECOND);
						throw e;
					}
				} while (packlen != recvlen - CommandLength.sizeof);
				
				tail = command_offset + packlen + CommandLength.sizeof;
				
				// キューに貯める // バグっているとのこと
				command = new NyRawCommand(m_buffer[command_offset .. tail].dup);
				m_command_queue.push(command);
				
				m_register.clear();
				return;
			}
			
			/* 受信長のほうが大きい場合	*/
			
			// キューに貯める
			command = new NyRawCommand(m_buffer[command_offset 
												.. command_offset + packlen + CommandLength.sizeof].dup);
			m_command_queue.push(command);
			
			// 次のコマンド
			command_offset 	+= packlen + CommandLength.sizeof;
			recvlen			-= packlen + CommandLength.sizeof;
		}
		
		m_register.clear();
	}
	
public:
	this()
	{
		m_recv_rc4key 	= new RC4;
		m_send_rc4key 	= new RC4;
		m_command_queue	= new Queue!(NyRawCommand);
		m_fdset 		= new SocketSet(1);
		
		ubyte *p = cast(ubyte *)malloc(BUFFER_SIZE);
//		writef("- %s malloc\n", s_malloc_queue++);
		
		if (!p) {
			throw new OutOfMemoryException;
		}
		m_buffer		= p[0 .. BUFFER_SIZE];
		m_buffer_ptr 	= p;
	}
	
	// 通信速度(B/S)
	uint rate()
	{
		time_t cur = time(null);
		uint rate;
		
		uint recv_diff = cur - m_last_recv_time;
		uint send_diff = cur - m_last_send_time;
		
		if (recv_diff == 0 && send_diff == 0) {
			rate = 0;
		} else if (recv_diff == 0) {
			rate = m_send_size_sec / send_diff;
		} else if (send_diff == 0) {
			rate = m_recv_size_sec / recv_diff;
		} else {
			rate = (m_send_size_sec / send_diff) + (m_recv_size_sec / recv_diff);
		}
		
		return rate;
	}
	
	// 接続時間(秒)
	uint ctl()
	{
		return cast(uint)(time(null) - m_start_time);
	}
	
	uint queueCount()
	{
		return m_command_queue.length;
	}
	
	InternetAddress address()
	{
		if (m_socket is null) {
			return null;
		}
		return cast(InternetAddress)m_socket.remoteAddress();
	}
	
	void close()
	{
		if (m_socket !is null) {
			m_socket.shutdown(SocketShutdown.SEND);
			m_socket.shutdown(SocketShutdown.RECEIVE);
			m_socket.close();
		}
		
		if (m_buffer_ptr != null) {
//			writef("- %s close free\n", s_malloc_queue--);
			
			free(m_buffer_ptr);
			m_buffer_ptr = null;
			m_buffer 	= null;
		}
		if (!m_command_queue.empty) {
//			writef("close: queue %s\n", m_command_queue.length);
			while (!m_command_queue.empty) {
				NyRawCommand command = m_command_queue.shift;
				command.clearBuffer();
				//command.destroy();
				command = null;
			}
		}
	}
	
	~this()
	{
		if (m_buffer_ptr != null) {
//			writef("- %s gc free\n", s_malloc_queue--);
			free(m_buffer_ptr);
			m_buffer_ptr = null;
		}
	}
	
	void timeout(uint sec, uint micro_sec)
	{
		m_timeval.seconds 		= sec;
		m_timeval.microseconds	= micro_sec;
	}
	
	void resetTimeout()
	{
		m_timeval.seconds		= m_timeval.seconds.init;
		m_timeval.microseconds	= m_timeval.microseconds.init;
	}
	
	socket_t handle()
	{
		if (m_socket !is null) {
			return m_socket.handle;
		}
		
		return socket_t.init;
	}
	
	void listen(ushort port, int backlog = 128)
	{
		InternetAddress localaddr = new InternetAddress(port);
		int[] optval = new int[1];
		
		optval[0] = 1;
		
		m_socket = new TcpSocket();
		m_socket.setOption(SocketOptionLevel.SOCKET, SocketOption.REUSEADDR, optval);
		m_socket.blocking(true);
		m_socket.bind(localaddr);
		m_socket.listen(backlog);
		m_inetaddr		= localaddr;
		m_is_listener 	= true;
		m_is_active		= true;
	}
	
	NyConnection accept(out NyProtocolHeader header)
	{
		ubyte[]			recv_init_block	= new ubyte[6];
		ubyte[]			send_init_block	= generateRandomBytes(6);
		NyConnection 	connection		= new NyConnection;
		Socket 			client_socket 	= m_socket.accept();
		InternetAddress	remote_address;
		uint			recvlen;
		
		remote_address = cast(InternetAddress)client_socket.remoteAddress;
		client_socket.blocking(false);
		
		connection.m_socket		= client_socket;
		connection.m_timeval	= m_timeval;
		connection.m_is_active	= true;
		connection.m_inetaddr	= new InternetAddress(remote_address.addr, remote_address.port);
		connection.m_start_time	= time(null);
		
		try {
//			writef("--accept\n");
			header = connection.authorize();
//			writef("--accept %s\n", header.toString);
		} catch (ConnectionTimeOut te) {
			// delete connection;
			// connection.close();
			throw te;
		} catch (Exception e) {
			// delete connection;
			// connection.close();
			throw e;
		}
		
		return connection;
	}
	
	NyProtocolHeader connect(InternetAddress connect_to)
	{
		ubyte[] 			init_block;
		NyProtocolHeader	header;
		int					recvlen, tryover;
		
		// 各メンバ初期化
		connectionInitialize();
		m_is_listener	= false;
		m_socket		= new TcpSocket();
		m_inetaddr		= new InternetAddress(connect_to.addr, connect_to.port);
		
		
		// コネクションポーリング
		int result;
		auto SocketSet wrtfd = new SocketSet(1);
		auto SocketSet errfd = new SocketSet(1);
		
		m_lasttimeval = m_timeval;
		
		wrtfd.add(m_socket);
		errfd.add(m_socket);
		m_socket.blocking(false); // non block socket
		
		m_socket.connect(connect_to); // 
		if (m_timeval.seconds == int.init && m_timeval.microseconds == int.init) {
			while ((result = Socket.select(null, wrtfd, errfd, null)) == -1)
				{} // interrupt
		} else {
			m_lasttimeval = m_timeval;
			while ((result = Socket.select(null, wrtfd, errfd, &m_lasttimeval)) == -1)
				{} // interrupt
		}
		
		switch (result) {
		case 0:		// タイムアウト
			throw new ConnectionTimeOut();
			
		default:	// OK?
			if (errfd.isSet(m_socket)) {
				throw new SocketException("接続エラー", result);
			}
			break; // connect was OK
		}
		m_is_active = true;
		m_start_time = time(null);
		
		// 接続開始
		return authorize();
	}
	
	enum Poll {
		READABLE = 0,
		WRITEABLE,
		
		__READABLE,	// 本当の話
	}
	
	void poll(Poll which)
	{
		int result;
		
		if (which == Poll.READABLE && !m_command_queue.empty) {
			return;
		}
		if (which == Poll.__READABLE) {
			which = Poll.READABLE;
		}
		
		// ポーリング
		m_fdset.reset();
		m_fdset.add(m_socket);
		
		if (which == Poll.READABLE) {
			if (m_timeval.seconds == int.init && m_timeval.microseconds == int.init) {
				result = Socket.select(m_fdset, null, null, null);
			} else {
				m_lasttimeval = m_timeval;
				result = Socket.select(m_fdset, null, null, &m_lasttimeval);
			}
		} else {// if WRITEABLE
			if (m_timeval.seconds == int.init && m_timeval.microseconds == int.init) {
				result = Socket.select(null, m_fdset, null, null);
			} else {
				m_lasttimeval = m_timeval;
				result = Socket.select(null, m_fdset, null, &m_lasttimeval);
			}
		}
		switch (result) {
		case 0:		// タイムアウト
			throw new ConnectionTimeOut();
		
		default:	// OK?
			if (!m_fdset.isSet(m_socket)) {
				throw new SocketException("ソケットエラー(select)", result);
			}
			break; // receive was OK
		}
	}
	
	// 1コマンドずつ受信
	NyRawCommand receiveOneCommand()
	{
		NyRawCommand command;
		if (! m_command_queue.empty) {
			return  m_command_queue.shift;
		}
		
		uint packet_length;
		int  recvlen;
		
		// レングス受信
		recvlen = 0;
		do {
			recvlen += receive(m_buffer[recvlen .. CommandLength.sizeof]);
		} while (recvlen != CommandLength.sizeof);
		
//		m_recv_rc4key.crypt(m_buffer[0 .. CommandLength.sizeof]);
		
		packet_length = *cast(uint *)m_buffer.ptr;
		if (packet_length > BLOCK_MAX) {
			throw new DecryptError(packet_length);
		}
		
		// コマンド受信
		recvlen = 0;
		while (recvlen != packet_length) {
			recvlen += receive(m_buffer[CommandLength.sizeof + recvlen .. CommandLength.sizeof + packet_length]);
		}
		if (recvlen == 0) {
			throw new DecryptError(0);
		}
		
		// 復号
//		m_recv_rc4key.crypt(m_buffer[CommandLength.sizeof .. CommandLength.sizeof + packet_length]);
		command = new NyRawCommand(m_buffer[0 .. CommandLength.sizeof + packet_length].dup);
		
		return command;
	}
	
	// バッファリング受信
	NyRawCommand receiveCommand()
	{
		NyRawCommand 	raw_command;
		
		if (m_command_queue.empty) {
			// キューが空の場合はコマンドを補充する
			try {
				replenish();
			} catch (SocketException e) {
				if (!m_command_queue.empty) {
					raw_command = m_command_queue.shift;
					return raw_command;
				}
				throw e;
			}
		}
		raw_command = m_command_queue.shift;
		
		return raw_command;
	}
	
	void sendCommand(NyCommand command)
	{
		ubyte[] packet = command.pack;
		
		try {
			send(packet);
		} finally {
			delete packet;
			packet = null;
		}
	}
	
	bool isActive()
	{
		return m_is_active ? true: false;
	}
	
	bool isListener()
	{
		return m_is_listener ? true: false;
	}
	
	uint upstreamRateMax()
	{
		return m_send_rate_max;
	}
	
	void upstreamRateMax(uint rate_max)
	{
		m_send_rate_max = rate_max;
	}
	
	uint downstreamRateMax()
	{
		return m_send_rate_max;
	}
	
	void downstreamRateMax(uint rate_max)
	{
		m_send_rate_max = rate_max;
	}
	
	void blocking(bool block)
	{
		m_socket.blocking = block;
	}
}

