﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.system.task;
/*
 * タスク
 * 
 *
 */
 
private import	std.thread;
private import  std.string;
private import	std.random;
private import	std.stream;
private import	std.date;

private import	pes.system.semaphore;
private import	pes.system.eventflag;
private import	pes.system.mailbox;
private import	pes.system.delay;
private import	pes.system.synchronizer;

private import  std.stdio;
private import	std.gc;
private import	std.random;
private import	std.c.time;



class TaskResource
{
protected:
	MailResource		m_mail;
	EventResource		m_event;
	SemaphoreResource	m_semaphore;
	
public:
	
	this()
	{
		m_mail			= new MailResource;
		m_event			= new EventResource;
		m_semaphore		= new SemaphoreResource;
	}
	
	MailResource mailBox()
	{
		return m_mail;
	}
	
	EventResource eventFlag()
	{
		return m_event;
	}
	
	SemaphoreResource semaphore()
	{
		return m_semaphore;
	}
}

version(Win32)
{
	private import std.c.windows.windows;
}

class TaskException: Exception
{
	this(char[] msg)
	{
		super(msg);
	}
}


class Task: Thread
{
protected:
	uint				m_gc_time;
	uint				m_gc_limit;
	int					m_pause_counter;
	TaskResource		m_resource;
	mailbox				m_mailbox;
	Synchronizer		m_synchronizer;
	
	abstract int mainLoop();
	override int run()
	{
		int result;
		
		result = mainLoop();
		
		return result;
	}
	
public:
	this(mailbox id, TaskResource resource)
	{
		m_mailbox 	= id;
		m_resource	= resource;
		m_gc_limit	= 32;
		m_gc_time	= rand() % gcLimit;
		m_synchronizer = new Synchronizer();
		
//		std.gc.enable;
	}
	
	Synchronizer synchronizer()
	{
		return m_synchronizer;
	}
	
	void fullCollect()
	{
		// メモリのことは分からないので、
		// 最低gc_time回以上たってから行う
		if (m_gc_time++ > gcLimit) {
//			log("start fullCollect");
//			std.gc.fullCollect();
//			log("end fullCollect");
			m_gc_time = 0;
		}
	}
	
	uint gcLimit()
	{
		return m_gc_limit;
	}
	
	void gcLimit(uint limit)
	{
		m_gc_limit = limit;
	}
	
	override void pause()
	{
		try {
			super.pause();
		} catch (ThreadError) {
			// do nothing
		}
	}
	

	override void resume()
	{
		try {
			super.resume();
		} catch (ThreadError) {
			// do nothing
		}
	}
	

/+
	// 自己サスペンドするため
	// 不整合が起こらないために.
	// true : サスペンド状態からレジュームされた,
	// false: ランニング中だった
	bool tryResume()
	{
		version(Win32) 
		{
			DWORD suspend_counter;
			
			if (getState != TS.RUNNING) {
				throw new ThreadError("cannot resume 1");
				return true;
			}
			
			suspend_counter = ResumeThread(hdl);
			if (suspend_counter == 0xFFFFFFFF) {
				throw new ThreadError("cannot resume 2");
			}
			
			return suspend_counter != 0;
		}
		
		version(linux)
		{
			assert(0); // not implemented
		}
	}
+/	
	TaskResource resource()
	{
		return m_resource;
	}
	
	mailbox mailboxId()
	{
		return m_mailbox;
	}
	
	// メッセージ関係
	bool pollMail(uint wait_msec = 0)
	{
		return m_resource.mailBox.poll(m_mailbox, wait_msec);
	}
	
	Mail receiveMail()
	{
		return m_resource.mailBox.receive(m_mailbox);
	}
	
	void sendMail(mailbox to, Mail mail)
	{
		mail.from = m_mailbox;
		m_resource.mailBox.send(to, mail);
	}
	
	// イベント関係
	void waitForStartEvent(eventflag id, uint mask)
	{
		m_resource.eventFlag.waitForStartEvent(id, mask);
	}
	
	void waitForEndEvent(eventflag id, uint mask)
	{
		m_resource.eventFlag.waitForEndEvent(id, mask);
	}
	
	bool pollStartEvent(eventflag id, uint mask, uint wait_msec = 0)
	{
		return m_resource.eventFlag.pollStartEvent(id, mask, wait_msec);
	}
	
	bool pollEndEvent(eventflag id, uint mask, uint wait_msec = 0)
	{
		return m_resource.eventFlag.pollEndEvent(id, mask, wait_msec);
	}
	
	void startEvent(eventflag id, uint mask)
	{
		m_resource.eventFlag.startEvent(id, mask);
	}
	
	void endEvent(eventflag id, uint mask)
	{
		m_resource.eventFlag.endEvent(id, mask);
	}
	
	uint referEvent(eventflag id)
	{
		return m_resource.eventFlag.referEvent(id);
	}
	
	// セマフォ
	void signalSemaphore(semaphore id)
	{
		m_resource.semaphore.signal(id);
	}
	
	void releaseSemaphore(semaphore id)
	{
		m_resource.semaphore.release(id);
	}
	
	static Task getThis()
	{
		Task task = cast(Task)Thread.getThis();
		
		if (task is null) {
			throw new TaskException(std.string.format("cast error %s",
											Thread.getThis.classinfo.name));
		}
		
		return task;
	}
	
	void log(char[] msg)
	{
		synchronized {
			debug(TaskBase) writef(derr, "\nLOG: %s: %s\n", this.toString, msg);
			
			try {
				File file = new File("task.log", FileMode.Append);
				
				file.writef("%s:%30s: %s\n",
					std.date.toString(UTCtoLocalTime(getUTCtime)), this.toString, msg);
				
				file.close();
				if (std.file.getSize("task.log") > 0x100000) {
					std.file.remove("task.log");
				}
			} catch (Object) { }
		}
	}
	
	void errorLog(char[] msg)
	{
		synchronized {
			debug(TaskBase) writef(derr, "\nERROR_LOG: %s: %s\n", this.toString, msg);
			
			try {
				File file = new File("dead.log", FileMode.Append);
				
				file.writef("%s:%30s: %s\n",
					std.date.toString(UTCtoLocalTime(getUTCtime)), this.toString, msg);
				
				file.close();
			} catch (Object) { }
		}
	}
}
