﻿module common.keytable;

private import std.string;
private import std.utf;
private import std.stdio;

private import pes.wstring;
private import pes.random;
private import pes.net.winny.key;
private import pes.system.locker;
private import pes.system.semaphore;
private import pes.data.sqlite3;

import std.stdio;

// キー要素
class KeyInformation
{
protected:
	NyKeyInformation	m_key;
	
public:
	NyKeyInformation key()
	{
		return m_key;
	}
	
	// ソート用
	override int opCmp(Object o)
	{
		KeyInformation k = cast(KeyInformation)o;
		
		return k.m_key.timer - m_key.timer;
	}
	
	this(NyKeyInformation k)
	{
		m_key 		= k;
	}
	
	ubyte[] aaKey()
	{
		return m_key.hash;
	}
}


const char[] KEYTABLE_CREATE = 
	"CREATE TEMP TABLE KEY_TABLE(hash blob, size int, name text, trip text, timer int, packet blob)";
	
//const char[] KEYTABLE_CREATE = 
//	"CREATE TABLE KEY_TABLE(hash blob, size int, name text, trip text, timer int, packet blob)";
	
const char[] KEYTABLE_DROP =
	"DROP TABLE KEY_TABLE";
	
const char[] KEYTABLE_CLEAR =
	"DELETE FROM KEY_TABLE";


const char[] KEYTABLE_CLEAR_COUNT =
"DELETE FROM KEY_TABLE WHERE hash IN (SELECT hash FROM KEY_TABLE ODER BY timer ASC LIMIT ?)";

const char[] KEYTABLE_HASH_INDEX_CREATE = 
	"CREATE UNIQUE INDEX KEY_TABLE_HASH_IDX on KEY_TABLE(hash)";

const char[] KEYTABLE_NAME_INDEX_CREATE = 
	"CREATE INDEX KEY_TABLE_NAME_IDX on KEY_TABLE(name)";

const char[] KEYTABLE_TTL_INDEX_CREATE = 
	"CREATE INDEX KEY_TABLE_TIMER_IDX on KEY_TABLE(timer)";

const char[] KEYTABLE_INSERT =
	"INSERT INTO KEY_TABLE (hash, size, name, trip, timer, packet) values(?, ?, ?, ?, ?, ?)";

const char[] KEYTABLE_EXISTS_DELETE =
	"DELETE FROM KEY_TABLE WHERE hash=?";

const char[] KEYTABLE_SERACH_HASH =
	"SELECT packet, timer FROM KEY_TABLE WHERE hash=? LIMIT ?";

const char[] KEYTABLE_SERACH_NAME =
	"SELECT packet, timer FROM KEY_TABLE WHERE name like('%'|| ? ||'%') LIMIT ?";

const char[] KEYTABLE_SERACH_TRIP =
	"SELECT packet, timer FROM KEY_TABLE WHERE trip = ? LIMIT ?";

const char[] KEYTABLE_TTL_DOWN = 
	"UPDATE KEY_TABLE SET timer = timer - 1";

const char[] KEYTABLE_TTL_DOWN30 = 
	"UPDATE KEY_TABLE SET timer = timer - 30";

const char[] KEYTABLE_TTL_DIE = 
	"DELETE FROM KEY_TABLE WHERE timer <= 0";

const char[] KEYTABLE_COUNT =
	"SELECT COUNT(hash) FROM KEY_TABLE";

const char[] KEYTABLE_DB_FILE = "keytable.db";

extern (C) int callback_func(void *NotUsed, int argc, char **argv, char **azColName){
  int i;
  for(i=0; i<argc; i++){
    printf("%s = %s\n", azColName[i], argv[i] ? argv[i] : "NULL");
  }
  printf("\n");
  return 0;
}



class KeyTable: Locker!(ubyte[], KeyInformation)
{
protected:
	SQLite m_db;
	
public:
	this(SemaphoreResource resource, semaphore id)
	{
		m_db = new SQLite(KEYTABLE_DB_FILE);
		
//		try {
//			m_db.exec(KEYTABLE_DROP);
//		} catch (SQLiteException) {
//			//
//		}
		m_db.exec(KEYTABLE_CREATE);
		m_db.exec(KEYTABLE_HASH_INDEX_CREATE);
		m_db.exec(KEYTABLE_TTL_INDEX_CREATE);
		m_db.exec(KEYTABLE_NAME_INDEX_CREATE);
		
		super(resource, id);
	}
	
	void close()
	{
		m_db.close();
	}
	
	void addKeyArray(NyKeyInformation[] array, LockerMode mode = LockerMode.AUTO)
	{
		KeyInformation key;
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			m_db.begin();
			
			//INSERT INTO KEY_TABLE(hash, size, name, trip, timer, packet) values(?, ?, ?, ?, ?, ?);
			SQLiteStatement insert_stmt = m_db.prepare(KEYTABLE_INSERT);
			SQLiteStatement delete_stmt = m_db.prepare(KEYTABLE_EXISTS_DELETE);
			
			foreach (NyKeyInformation e; array) {
				delete_stmt.bind(1, e.hash);
				delete_stmt.step();
				
				insert_stmt.bind(1, e.hash, e.fileSize, e.fileName, e.trip, e.timer, e.pack());
				insert_stmt.step();
				
				delete_stmt.reset();
				insert_stmt.reset();
			}
			delete_stmt.finalize();
			insert_stmt.finalize();
			
			m_db.commit();
		}
	}
	
	NyKeyInformation[] searchHash(ubyte[] hash, uint max = uint.max, LockerMode mode = LockerMode.AUTO)
	{
		NyKeyInformation[] result;
		uint i, count;
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			SQLiteStatement stmt = m_db.prepare(KEYTABLE_SERACH_HASH);
			
			stmt.bind(1, hash, max);
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {
				NyKeyInformation key 	= new NyKeyInformation;
				ubyte[] packet 			= stmt.fetch!(ubyte[])(0);
				uint	timer			= stmt.fetch!(int)(1);
				
				if (timer <= 0) {
					continue;
				}
				key.unpack(packet);
				key.timer = timer;
				result ~= key;
				if (++i >= max) {
					break;
				}
			}
			stmt.finalize();
		}
		
		return result;
	}
	
	NyKeyInformation[] searchName(char[] keyword, uint max = uint.max, LockerMode mode = LockerMode.AUTO)
	{
		NyKeyInformation[] result;
		uint i, count;
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			SQLiteStatement stmt = m_db.prepare(KEYTABLE_SERACH_NAME);
			
			stmt.bind(1, keyword, max);
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {
				NyKeyInformation key 	= new NyKeyInformation;
				ubyte[] packet 			= stmt.fetch!(ubyte[])(0);
				int		timer			= stmt.fetch!(int)(1);
				
				if (timer <= 0) {
					continue;
				}
				
				key.unpack(packet);
				key.timer = timer;
				result ~= key;
				if (++i >= max) {
					break;
				}
			}
			stmt.finalize();
		}
		
		return result;
	}
	
	NyKeyInformation[] searchTrip(char[] trip, uint max = uint.max, LockerMode mode = LockerMode.AUTO)
	{
		NyKeyInformation[] result;
		uint i, count;
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			SQLiteStatement stmt = m_db.prepare(KEYTABLE_SERACH_TRIP);
			
			stmt.bind(1, trip, max);
			stmt.step();
			count = stmt.dataCount;
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {
				NyKeyInformation key 	= new NyKeyInformation;
				ubyte[] packet 			= stmt.fetch!(ubyte[])(0);
				int		timer			= stmt.fetch!(int)(1);
				
				if (timer <= 0) {
					continue;
				}
				
				key.unpack(packet);
				key.timer = timer;
				result ~= key;
				if (++i >= max) {
					break;
				}
			}
			stmt.finalize();
		}
		
		return result;
	}
	
	override uint length(LockerMode mode = LockerMode.AUTO)
	{
		uint count = 0;
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		synchronized(lock) {
			SQLiteStatement stmt = m_db.prepare(KEYTABLE_COUNT);
			stmt.step();
			if (stmt.hasRow) {
				count = stmt.fetch!(int)(0);
			}
			stmt.finalize();
		}
		
		return count;
	}
	
	void clear(uint clear_count, LockerMode mode = LockerMode.AUTO)
	{
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		
		synchronized(lock) {
			SQLiteStatement stmt = m_db.prepare(KEYTABLE_CLEAR_COUNT);
			stmt.bind(1, clear_count);
			stmt.step();
			stmt.finalize();
		}
	}
	
	override void clear(LockerMode mode = LockerMode.AUTO)
	{
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		synchronized(lock) {
			m_db.begin();
			m_db.exec(KEYTABLE_CLEAR);
			m_db.commit();
		}
	}
	
	void tick(LockerMode mode = LockerMode.AUTO)
	{
		Object lock;
		
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		synchronized(lock) {
			m_db.begin();
			m_db.exec(KEYTABLE_TTL_DOWN);
			m_db.exec(KEYTABLE_TTL_DIE);
			m_db.commit();
		}
	}
	
	void tick30(LockerMode mode = LockerMode.AUTO)
	{
		Object lock;
		if (mode == LockerMode.AUTO) {
			lock = lockObject;
		} else {
			lock = new Object;
		}
		synchronized(lock) {
			m_db.begin();
			m_db.exec(KEYTABLE_TTL_DOWN30);
			m_db.exec(KEYTABLE_TTL_DIE);
			m_db.commit();
		}
	}
}
